#!/bin/perl
#
#
# COURTNEY: program to detect SATAN scans.
#
# Copyright (c) 1995
#   The Regents of the University of California.  All rights reserved.
#
# This work was produced at the University of California, Lawrence
# Livermore National Laboratory (UC LLNL) under contract number
# W-7405-ENG-48 (Contract 48) between the U.S. Department of Energy
# (DOE) and the Regents of the University of California (University) for
# the operation of UC LLNL.  Copyright is reserved to the University for
# purposes of controlled dissemination, commercialization through formal
# licensing, or other disposition under Contract 48; DOE policies,
# regulations and orders; and U.S. statutes.  The rights of the Federal
# Government are reserved under Contract 48 subject to the restrictions
# agreed upon by the DOE and University as allowed under DOE Acquisition
# Letter 88-1.
#
# REVISIONS:
#   3/22/95: Marvin J. Christensen  V1.0  Original release 
#   3/29/95: Paul S. Mauvais 	    V1.01 Changed MONITOR to false to
#					  save screen traffic
#   3/29/95: Steve Weeber           V1.02 Added syslog reporting
#
# ----------------------------------------------------------------------

require "syslog.pl";

@assoc_lists = ( \%sunrpc,  \%icmp,    \%ttime,   \%telnet,   \%smtp,
                 \%ftp,     \%whois,   \%domain,  \%gopher,   \%www,
                 \%finger,  \%exec,    \%login,   \%shell,    \%printer,
                 \%uucp,    \%tcpmux,  \%echo,    \%discard,  \%systat,
                 \%daytime, \%netstat, \%chargen, \%tftp,     \%name,
                 \%biff,    \%syslog,  \%talk,    \%portscan, \%xwindows );

@assoc_list  = (  'sunrpc',  'icmp',    'ttime',   'telnet',   'smtp', 
                  'ftp',     'whois',   'domain',  'gopher',   'www',
                  'finger',  'exec',    'login',   'shell',    'printer',
                  'uucp',    'tcpmux',  'echo',    'discard',  'systat',
                  'daytime', 'netstat', 'chargen', 'tftp',     'name',
                  'biff',    'syslog',  'talk',    'portscan', 'xwindows' );

$UPDATE_INTERVAL = 5;		# update host information every 
				# X minutes
$OLD_AGE = 7;			# get rid of hosts that are
				# older that X minutes.

$HIGH_THRESHOLD = 15;		# heavy  "SATAN" attack
$LOW_THRESHOLD  = 9;		# normal "SATAN" attack

$TRUE      = 1;
$FALSE     = 0;
$last_time = 0;
$MODULUS   = 1440;		# number of minutes per day

$DEBUG     = $FALSE;		# major verbose
$MONITOR   = $FALSE;		# prints counted host connections

$SYSLOG	   = $TRUE;		# send messages to syslog
$SCREEN    = $FALSE;		# send messages to stdout


#
# Starting loop of main program
#

open (TCPDUMP, "tcpdump -l '\
    (icmp[0] == 8 ) or \
    port sunrpc or \
    port (1 or 10 or 100 or 1000 or 5000 or 10000 or 20000 or 30000) or \
    (port (6000 or 6001 or 6002 or 6010 or 6011 or 6012) and \
     (tcp[13] &2 != 0) )  or \
    (port (tcpmux or \
	   echo or \
	   discard or \
	   systat or \
	   daytime or \
	   netstat or \
	   chargen or \
	   ftp or \
	   telnet or \
	   smtp or \
	   time or \
	   whois or \
	   domain or\
	   70 or \
	   80 or \
	   finger or \
	   tftp or \
	   login or \
	   uucp or \
	   printer or \
	   shell or \
	   exec or \
	   name or \
	   biff or \
	   syslog or \
	   talk) and \
     (tcp[13] &2 != 0) )  \
    ' |") || die "Unable to execute tcpdump command: $!\n";



while(<TCPDUMP>) {

 chop;
 @line =  split;

#
# parse the input line for time, source, dest, and protocol
#

  if($line[4] =~ "icmp") {
    ($time)        = $line[0] =~ /(\d+:\d+:\d+).\d+/;
    ($src)         = $line[1];
    ($dst)         = $line[3]; 
                     chop($dst);
    ($proto)       = $line[4];
                     chop($proto);
  } else {
    ($time)        = $line[0] =~ /(\d+:\d+:\d+).\d+/;
            @src_a = split(/\./, $line[1]);
    ($src)         = join(".", @src_a[0..($#src_a - 1)] );
            @dst_a = split(/\./, $line[3]);
    ($dst)         = join(".", @dst_a[0..($#dst_a - 1)] );
    ($proto)       = @dst_a[$#dst_a];
                     chop($proto);
  }


#
# calculate time based on input 
#  only worry about minutes
#  modulus 1440 (24 hours * 60 minutes/hour)
#  update_time_interval based on UPDATE_INTERVAL


   ($hours, $minutes, $junk) = split(/:/, $time);
   $current_time = ((60 * $hours) + $minutes);
   
   $delta_time = ($current_time >= $last_time) ? 
                     ($current_time - $last_time) : 
                    (($current_time + $last_time) % $MODULUS);
   if ($delta_time >= $UPDATE_INTERVAL ) {
      $time_to_update = $TRUE;
      $last_time = $current_time;
   } else {
      $time_to_update = $FALSE;
   }


#
# create associative arrays based on protocol
#

  ENTRY: {

     if($proto eq "sunrpc")  { $sunrpc{$src}   = $current_time; last ENTRY };
     if($proto eq "icmp")    { $icmp{$src}     = $current_time; last ENTRY };
     if($proto eq "time")    { $ttime{$src}    = $current_time; last ENTRY }; 
     if($proto eq "telnet")  { $telnet{$src}   = $current_time; last ENTRY }; 
     if($proto eq "smtp")    { $smtp{$src}     = $current_time; last ENTRY }; 
     if($proto eq "ftp")     { $ftp{$src}      = $current_time; last ENTRY }; 
     if($proto eq "whois")   { $whois{$src}    = $current_time; last ENTRY }; 
     if($proto eq "domain")  { $domain{$src}   = $current_time; last ENTRY }; 
     if($proto eq "70")      { $gopher{$src}   = $current_time; last ENTRY }; 
     if($proto eq "80")      { $www{$src}      = $current_time; last ENTRY }; 
     if($proto eq "finger")  { $finger{$src}   = $current_time; last ENTRY }; 
     if($proto eq "exec")    { $exec{$src}     = $current_time; last ENTRY }; 
     if($proto eq "login")   { $login{$src}    = $current_time; last ENTRY }; 
     if($proto eq "shell")   { $shell{$src}    = $current_time; last ENTRY }; 
     if($proto eq "printer") { $printer{$src}  = $current_time; last ENTRY }; 
     if($proto eq "uucp")    { $uucp{$src}     = $current_time; last ENTRY }; 
     if($proto eq "tcpmux")  { $tcpmux{$src}   = $current_time; last ENTRY }; 
     if($proto eq "echo")    { $echo{$src}     = $current_time; last ENTRY }; 
     if($proto eq "discard") { $discard{$src}  = $current_time; last ENTRY }; 
     if($proto eq "systat")  { $systat{$src}   = $current_time; last ENTRY }; 
     if($proto eq "daytime") { $daytime{$src}  = $current_time; last ENTRY }; 
     if($proto eq "netstat") { $netstat{$src}  = $current_time; last ENTRY }; 
     if($proto eq "chargen") { $chargen{$src}  = $current_time; last ENTRY }; 
     if($proto eq "tftp")    { $tftp{$src}     = $current_time; last ENTRY }; 
     if($proto eq "name")    { $name{$src}     = $current_time; last ENTRY }; 
     if($proto eq "biff")    { $biff{$src}     = $current_time; last ENTRY }; 
     if($proto eq "syslog")  { $syslog{$src}   = $current_time; last ENTRY }; 
     if($proto eq "talk")    { $talk{$src}     = $current_time; last ENTRY }; 
     if( ($proto == 1)     || ($proto ==10 )    || ($proto == 100)    || 
         ($proto == 1000)  || ($proto == 5000 ) || ($proto == 10000)  ||
         ($proto == 20000) || ($proto == 30000) )  
                             { $portscan{$src} = $current_time; last ENTRY }; 
    if( ($proto == 6000)  || ($proto == 6001)  || ($proto == 6002)   ||
         ($proto == 6010)  || ($proto == 6011)  || ($proto == 6012) )  
                             { $xwindows{$src} = $current_time; last ENTRY }; 

      print "PROTO: $proto, \n"  if $DEBUG;

  }


#
# clean up associative array list getting rid of entries where the
# time (e.g., value) is older than $OLD_AGE
#

  if ($time_to_update) {
    print "TIME_TO_UPDATE $current_time, $last_time \n" if $DEBUG;

    undef @heavy_attack;                         # clean up lists that
    undef @normal_attack;                        # have identified hosts

    $entry = 0;                                  # used for debug
    foreach $assoc_name (@assoc_lists) {

      print "ENTRY $entry  $assoc_list[$entry] \n" if $DEBUG;
      $entry++;

      local(*ary) = $assoc_name;                 # dereference pointer
                                                 # don't know why this works
      foreach $key (sort keys %ary ) {        
        print "foreach delete: $key $ary{$key} <-> $current_time \n" if $DEBUG;

#        print "delete: $ary{$key}\n" if (( ($current_time >= $ary{$key}) ? 
#                                ($current_time -  $ary{$key}) : 
#                                (($current_time + $ary{$key}) % $MODULUS) )
#                                >= $OLD_AGE);
        delete($ary{$key}) if (( ($current_time >= $ary{$key}) ? 
                                ($current_time -  $ary{$key}) : 
                                (($current_time + $ary{$key}) % $MODULUS) )
                                >= $OLD_AGE);
      };
    };
  };


#
# count the number of times a host is contained in the associative
# arrays. For each assocative array, extract the hostname and
# increment the count for that hostname.
#
  undef %host_count;
  $entry = 0;                                     # used for debug
  foreach $assoc_name (@assoc_lists) {
   print "ENTRY $entry $assoc_list[$entry] \n" if $DEBUG;
   $entry++;

   local(*ary) = $assoc_name;

     foreach $key (sort keys %ary ) {
       print "hostcount: $key $ary{$key} \n" if $DEBUG;

       $host_count{ $key }++;

     };
  };


#
# If the host count is above a threshold, print out the hostname,
# using the timestamp from the tcpdump input, and then add the
# host to the array so that you don't keep printing the host
# once it has been identified. There are two lists, one for
# normal scans and one for heavy scans.

  foreach $host (keys %host_count) {
    print "counted: $host_count{$host} for $host \n" if ($DEBUG || $MONITOR);

    if ($host_count{$host} >= $HIGH_THRESHOLD) { 
      if (! $heavy_attack[$host]) {
        push(@heavy_attack, $host);
        undef $normal_attack[$host];
	if ($SYSLOG) {
	    &syslog ('alert', "courtney[$$]: HEAVY_ATTACK from $host\n");
	}
	if ($SCREEN) {
	    print "$time: HEAVY_ATTACK from $host\n";
	}
        print "$host added to heavy attack \n" if $DEBUG;
      };
    } 
    elsif ($host_count{$host} >= $LOW_THRESHOLD) {
     if (! $normal_attack[$host]) {
        push(@normal_attack, $host);
	if ($SYSLOG) {
	    &syslog ('alert', "courtney[$$]: NORMAL_ATTACK from $host\n");
	}
	if ($SCREEN) {
	    print "$time: NORMAL_ATTACK from $host\n";
	}
        print "$host added to normal attack \n" if $DEBUG;
      };
    };
  };
  

}; # endwhile

# all done. 
#
# TODO: 
#
# USE ONLY THE FIRST TCP RECORD.
# tcpdump generates two records for each establish connection. since
# courtney is only concerned with the originating machine, the second
# record is not used. However, the current code executes all the way
# through with this second record. courtney should stop right after the 
# switch statement and return back to the top of the while loop.



